' Todo - sub to accept, split and display a complete ASCII character
' Todo - code for Read from LCD
' Todo - allow 8 data lines
' Todo - split to 3 (?) documents.
' Todo - sub LCDHelp to display all subs and parameters
' Todo - establish naming conventions for parameters, subs
' Todo - doco.  Include reference to tutorials.
' Todo - test with all types of LCD 
' Todo - add test to check version >= 3.1



'
' demonstration program
' send lines to the LCD
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' LCD driver for MMBasic 3.1 or later (uses defined subroutines)
' Geoff Graham, Jan 2012.
' Modified by James Deakins to include 20 x 4 LCDs, Feb 2012.
'
' Will drive standard 16 x 2 or 20 x 4 LCDs
' For example:  futurlec.com       LCD16X2
'               altronics.com.au   Z7001
'               jaycar.com.au      QP5512
'                                  HD44780 (20 x 4)
'
' To use:
'  - Optionally, redefine the pins used with the command: ChangeLCDpins RS, EN, D1, D2, D3, D4, RW 
'  - Define LCD size with the command:  InitLCDVariables columns, lines
'  - Setup the LCD with the command:    InitLCD
'  - Display a line using the command:  PrintLCD LineNbr, Text$
'
' See documentation (LCD.PDF) for the schematic
' Maximite pin 11 is RS, pin 12 is EN
' pins 13 to 16 are D4 to D7
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
 ' following should go to the initialisation file, for insertion at the beginning of code
 ' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
 
DIM C_LineStart(4)     ' an array to hold the start addresses for each line
C_Lines = 0            ' variable to hold the number of lines on the LCD
C_LineLength = 0       ' variable to hold the line length of the LCD
RWenabled = 0			' variable indicating whether the Read Write pin is enabled (1=True)
 
' Allocate names to pin numbers based on LCD.PDF and the comments above.  
' The pin numbers may be changed by the ChangeLCDPins sub.

RS = 11
EN = 12
D1 = 13
D2 = 14
D3 = 15
D4 = 16
RW = 0	' Read Write - enable with sub LCDenableRW. Set with ChangeLCDPins sub.
'
' **** Change the parameters to show whether you have a 16 x 2, or a 20 x 4 LCD screen.
InitLCDSize(20, 4)

' if you need to change pin numbers, do it before calling InitLCD
'eg ChangeLCDPins(5, 6, 7, 8, 9, 10)	' - don't use unless you need to

InitLCD	' will need parameter for 4 or 8 bit

' Following should go to demo/doco file 
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

' display some simple lines, one at a time

PrintLCD 1, "  Hello  World"
PrintLCD 2, "  Maximite LCD"
PrintLCD 3, "with a 20 x 4 LCD," 
PrintLCD 4, "using Subroutines."
 
PAUSE 2000
' and here's a test of one of the subroutines...
TestMsg$ = "This is a really long message that we will scroll across the second line of the LCD screen."
ScrollLCDMessage(2, TestMsg$)

PAUSE 2000
' GOTO a screen location on the LCD and write a character.
'---------------------------------------------------------
LCDLocate(2,4) 		' GOTO L2, C4
c = ASC("Z")
LCD_Nibble INT(c/16), 1 : LCD_Nibble c, 1

' remainder (all subs) shoul go to the .lib file
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the size based variables for this application
' (C_Prefix indicates constants).
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCDSize(LCD_LineLength, LCD_Lines)

    C_LineLength = LCD_LineLength
    C_Lines = LCD_Lines
    '
    C_LineStart(1) = 0        ' Start address for line 1
    C_LineStart(2) = 64       ' Start address for line 2
    C_LineStart(3) = 20       ' Start address for line 3
    C_LineStart(4) = 84       ' Start address for line 4
   
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Change the Maximite / DM pins allocate for LCD control and data.
' argument #1 is the pin to use for the LCD RS line
' argument #2 is the pin to use for the LCD EN line
' argument #3 is the pin to use for the LCD D1 line
' argument #4 is the pin to use for the LCD D2 line
' argument #5 is the pin to use for the LCD D3 line
' argument #6 is the pin to use for the LCD D4 line
' argument #7 is the pin to use for the LCD BF line (optional)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB ChangeLCDPins(_RS, _EN, _D1, _D2, _D3, _D4, _RW)
    RS = _RS
    EN = _EN
    D1 = _D1
    D2 = _D2
    D3 = _D3
    D4 = _D4   
    IF _RW <> 0 Then 
    	IF RWenabled = 1 THEN	' true
    		RW = _RW 
    	ELSE
    		? "You must allow use of the Busy Flag using LCDenableRW"
    	ENDIF
    ENDIF
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the LCD hardware
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCD
	SETPIN RS, 9								' open collector
	SETPIN EN, 9								' open collector
	SETPIN D1, 9								' open collector
	SETPIN D2, 9								' open collector
	SETPIN D3, 9								' open collector
	SETPIN D4, 9								' open collector
	IF RWenabled THEN SETPIN RW, 9				' open collector (optional)
	
    LCD_Nibble &B0011, 0, 5                    ' reset
    LCD_Nibble &B0011, 0, 5                    ' reset
    LCD_Nibble &B0011, 0, 5                    ' reset
    LCD_Nibble &B0010, 0, 2                    ' 4 bit mode
    LCD_Nibble &B0010 : LCD_Nibble &B1100      ' 4 bits, 2 lines
    LCD_Nibble &B0000 : LCD_Nibble &B1100      ' display on, no cursor
    LCD_Nibble &B0000 : LCD_Nibble &B0110      ' increment on write
    LCD_Nibble &B0000 : LCD_Nibble &B0001, 0, 2 ' clear the display, pause 2
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Display a line on the LCD
' argument #1 is the line to be used (1 to 4)
' argument #2 is the line to display (can be any length)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB PrintLCD ( LineNumber, Line$ )
	LOCAL i, c, fillSpaces, fullLine$, curPosn

    curPosn = 128 + C_LineStart(LineNumber)

    ' first, send the cursor position (in two nibbles)
    LCD_Nibble INT(curPosn/16) : LCD_Nibble curPosn

    fillSpaces = C_LineLength - LEN(Line$) ' is the line full?
	IF fillSpaces > 0 THEN
    	fullLine$ = Line$ + SPACE$(fillSpaces)
    ELSE
    	fullLine$ = Line$
    ENDIF

    ' then send the text character by character (two nibbles per character)
    FOR i = 1 TO C_LineLength
    	c = ASC(MID$(fullLine$, i, 1))
        LCD_Nibble INT(c/16), 1 : LCD_Nibble c, 1
    NEXT i
    
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Send the lower 4 bits (called a nibble) to the LCD
' argument #1 is the nibble to send
' argument #2 is true if data, false if command (default is command)
' argument #3 is delay after the data has been sent (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble ( Data, Flag, Wait_mSec )
  PIN(RS) = Flag
  PIN(D1) = Data AND &B00000001
  PIN(D2) = Data AND &B00000010
  PIN(D3) = Data AND &B00000100
  PIN(D4) = Data AND &B00001000
  PIN(EN) = 1 : PIN(EN) = 0
  IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Move cursor to a location on the LCD screen.
' argument #1 is the line to be used (1 to maximum line number)
' argument #2 is the column to be used (1 to line length)  
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDLocate(Line, Column)
	LOCAL controlByte

    controlByte = 128 + C_LineStart(Line) + Column - 1
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Scroll a message across the LCD screen
' argument #1 is the line to be used (1 to 4)
' argument #2 is the message to display
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

SUB ScrollLCDMessage(line, Message$)
	LOCAL MsgLen, Mask_Offset, i, c

    MsgLen = LEN(Message$)
    FOR Mask_Offset = 0 TO MsgLen - C_LineLength ' keep track of where we're starting the display window
    	LCDLocate(line,1)                     ' go to the start of line
		FOR i = 1 TO C_LineLength
        	c = ASC(MID$(Message$, Mask_Offset + i, 1))
           	LCD_Nibble INT(c/16), 1 : LCD_Nibble c, 1
		NEXT i
        IF Mask_offset = 0 THEN 
        	PAUSE 1000	' initial - so user has time to read the initial part of the message
        ELSE
            PAUSE 400	' each char - so screen has time to refresh, and user can read it
        ENDIF
	NEXT Mask_Offset

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Enable the Read Write Flag.
' Most people won't use this, but it's here for completeness and
' so the Busy Flag and address can be read from the LCD
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDenableRW

	RWenabled = 1	' set to true
	
END SUB


'****************************************************************
' The following subroutines implement the LCD controls.
'****************************************************************

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Clear the LCD screen.
' Clears entire display and sets DDRAM address 0 in address counter
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDClearScreen
	LOCAL controlByte

	controlByte = &B00000001
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set cursor to Home position.
' Sets DDRAM address 0 in address counter.  Also returns display
' from being shifted to original position. DDRAM contents remain 
' unchanged.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDHome
	LOCAL controlByte

	controlByte = &B00000010
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Sets cursor move direction and specifies display shift.
' These operations are performed during data write and read.
' argument #1 is the move direction ("I" (Increment) or "D" (Decrement))
' argument #2 is the display shift ("On" or "Off")
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDEntryMode(MoveDir$, Shift$)
	LOCAL controlByte

	controlByte = &B00000100
	
    IF UCASE$(MoveDir$) = "I" THEN 
    	controlByte = controlByte AND &B00000010
    ENDIF
    
    IF UCASE$(Shift$) = "ON" THEN
    	controlByte = controlByte AND &B00000001
    ENDIF
       
    ? Bin$(controlByte)      
    
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
   
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set LCD display, cursor, and cursor blink controls 
' argument #1 is the Display      status ("On" or "Off")
' argument #2 is the Cursor       status ("On" or "Off")
' argument #3 is the Cursor Blink status ("On" or "Off")  
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDControl(Display$, Cursor$, Blink$)
	LOCAL controlByte

	controlByte = &B00001000
	
    IF UCASE$(Display$) = "ON" THEN 
    	controlByte = controlByte AND &B00000100
    ENDIF
    
    IF UCASE$(Cursor$) = "ON" THEN
    	controlByte = controlByte AND &B00000010
    ENDIF
   
    IF UCASE$(Blink$) = "ON" THEN
    	controlByte = controlByte AND &B00000001
    ENDIF 
     
    ? Bin$(controlByte)      
    
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
' &B00001111 is Display on, Cursor on, blinking cursor    
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Cursor and Display shift.
' Moves cursor and shifts display without changing DDRAM contents.
' argument #1 is the type ("S" (display Shift) or "C" (Cursor move)
' argument #2 is the shift direction ("L" (Left) or "R" (Right))
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDShift(Type$, Direction$)
	LOCAL controlByte

	controlByte = &B00010000
	
    IF UCASE$(Type$) = "S" THEN 
    	controlByte = controlByte AND &B00001000
    ENDIF
    
    IF UCASE$(Direction$) = "R" THEN
    	controlByte = controlByte AND &B00000100
    ENDIF
       
    ? Bin$(controlByte)      
    
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
   
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the interface data length, number of display lines and 
' character font.
' argument #1 is the number of bits used by the interface (4 or 8)
' note: this  code is currently only set up to use 4 bits
' argument #2 is the number of lines (1 or 2)
' TODO: Update with how to select 4 lines
' argument #3 is the font size ("L" (Large) or "N" (Normal))
' Large is 5 x 10 dots, Normal is 5 x 8 dots.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDFunctionSet(_Bits, _Lines, _Font$)
	LOCAL controlByte

	controlByte = &B00100000
	
    IF _Bits = 8 THEN 
    	? "Warning: 8 bit mode not currently supported!"
    	controlByte = controlByte AND &B00010000
    	EXIT SUB
    ENDIF
    
    IF _Lines = 2 THEN
    	controlByte = controlByte AND &B00001000
    ENDIF

    IF UCASE$(Type$) = "L" THEN 
    	controlByte = controlByte AND &B00000100
    ENDIF
         
    ? Bin$(controlByte)      
    
    LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
   
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' CGRAM data is sent and received after this setting.
' argument #1 is the CGRAM address (6 bits, so 0 to 63)
'
' In the character generator RAM, the user can rewrite character 
' patterns by program. For 5  8 dots, eight character patterns 
' can be written, and for 5  10 dots, four character patterns 
' can be written.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetCGRAMaddr(_Addr)
	LOCAL controlByte

	controlByte = &B01000000
	
	IF _Addr > 63 THEN
		? "CGRAM Address "; _Addr; " is too high (should be 0 to 63)"
	ELSE
    	controlByte = controlByte AND _Addr
    	? Bin$(controlByte)      
    	LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
    ENDIF	

    END SUB
    
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' DDRAM data is sent and received after this setting.
' argument #1 is the DDRAM address (7 bits, so 0 to 127)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetDDRAMaddr(_Addr)
	LOCAL controlByte

	controlByte = &B10000000
	
	IF _Addr > 127 THEN
		? "DDRAM Address "; _Addr; " is too high (should be 0 to 127)"
	ELSE
    	controlByte = controlByte AND _Addr
    	? Bin$(controlByte)      
    	LCD_Nibble INT(controlByte/16) : LCD_Nibble controlByte
    ENDIF	

END SUB 

    
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read busy flag. 
' Returns the Busy Flag (BF) indicating internal operation is being 
' performed and returns address counter contents.
' bf=1 - still internally operating. b=0 - instructions acceptable
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDreadBusyFlag(_Status, _Addr)
	LOCAL controlByte
	
' still under development
' pin(rw) = 1: pin(rw) = 0
' read 2 nibbles and reassemble to byte
'_Status = ret value AND 1000 0000 but return meaningful value
' _Addr  = ret value AND 0111 1111

END SUB 

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read the lower 4 bits (called a nibble) from the LCD
' argument #1 returns the nibble
' argument #2 is true if data, false if command (default is command)
' argument #3 is delay after the data has been sent (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble_Read ( Data, Flag, Wait_mSec )
	LOCAL Value
	' still under development
  	PIN(RS) = Flag
  	PIN(EN) = 1 : PIN(EN) = 0
  	IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
	? "Bits read are: ; PIN(D1); " "; PIN(D2); " "; PIN(D3); " "; PIN(D4)
  	Value = PIN(D1)	' low order bit
  	Value = Value OR (PIN(D2) * 2)
  	Value = Value OR (PIN(D3) * 4)
  	Value = Value OR (PIN(D4) * 8)

  	? Bin$(Value)
  	
  	Data = Value

END SUB